#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/* Misc utilities */

typedef enum { false, true } Flag;

void
error (const char *format, ...)
{
  va_list args;
  va_start (args, format);
  vfprintf (stderr, format, args);
  va_end (args);
  fprintf (stderr, "\n");
  exit (1);
}


void
stdlib_error (const char *context)
{
  error ("%s: %s", context, strerror (errno));
}


FILE *
open_file (const char *filename, const char *mode)
{
  if (strcmp (filename, "-") == 0)
    {
      assert (mode [0] == 'r' || mode [0] == 'w' || mode [0] == 'a');
      return mode [0] == 'r' ? stdin : stdout;
    }
  else
    {
      FILE *result = fopen (filename, mode);
      if (!result)
	stdlib_error (filename);
      return result;
    }
}


/* Elements */

typedef enum { R, G, B_color, C, M, Y, K, W, Root_color } Color;

enum { Root_size = 10 };

typedef struct
{
  unsigned int b: 1;
  unsigned int em: 1;
  unsigned int i: 1;
  unsigned int s: 1;
  unsigned int tt: 1;
  unsigned int u: 2;
  unsigned int size: 4;
  unsigned int color: 4;
  unsigned int character: 8;	/* belongs to Element but here for packing */
} 
Context;

typedef Context Element;


static Context
get_context (Element e)
{
  return (Context) e;
}


static Element
make_element (Context c, char ch)
{
  Element result = (Element) c;
  result.character = ch;
  return result;
}


static Flag
equal_contexts (Context c1, Context c2)
{
  return c1.b == c2.b 
    && c1.em == c2.em
    && c1.i == c2.i
    && c1.s == c2.s
    & c1.tt == c2.tt
    && c1.u == c2.u
    && c1.size == c2.size
    && c1.color == c2.color;
}


static void
check_context (Context context)
{
  assert (context.size <= Root_size);
  assert (R <= context.color && context.color <= Root_color);
}


static void
check_element (Element e)
{
  check_context (get_context (e));
  assert (isprint (e.character) || 
	  e.character == '\t' ||
	  e.character == '\r' ||
	  e.character == '\n');
}


static Context
make_root_context (void)
{
  Context result;
  result.b = result.em = result.i = result.s = result.tt = result.u = 0;
  result.size = Root_size;
  result.color = Root_color;
  check_context (result);
  return result;
}


/* Buffers */

enum { buffer_limit = 5 * 1024 * 1024 }; /* note 5 mega-character limit */

typedef struct 
{
  int ptr;
  Element elements[buffer_limit];
}
Buffer;


static void 
append (Buffer *buffer, Context context, char ch)
{
  check_context (context);
  if (buffer_limit <= buffer->ptr )
    error ("Input too big");
  buffer->elements[buffer->ptr] = make_element (context, ch);
  buffer->ptr++;
}


/* Return true iff the previous output character in buffer was a SPC
   decorated with the same space context as sc. */
static Flag
was_whitespace (const Buffer *buffer, Context sc)
{
  check_context (sc);
  if (0 == buffer->ptr)
    return false;
  else
    {
      const Element *e = buffer->elements + buffer->ptr - 1;
      return e->character == ' ' && equal_contexts (get_context (*e), sc);
    }
}


static Flag
equal (const Buffer *b1, const Buffer *b2)
{
  if (b1->ptr != b2->ptr)
    return 0;
  return
    0 == memcmp (b1->elements, b2->elements, b1->ptr * sizeof b1->elements[0]);
}


/* Canonicalization */

static Context
space_context (Context c)
{
  Context r = c;
  r.s = r.em = r.i = r.b = 0;
  if (r.u == 0)
    r.color = W;
  check_context (r);
  return r;
}


static void
append_whitespace_char (Buffer *buffer, Context context, char ch)
{
  Context sc = space_context (context);
  if (context.tt)
    append (buffer, sc, ch);
  else if (was_whitespace (buffer, sc))
    ;
  else
    append (buffer, sc, ' ');
}


/* Parsing */

static void
syntax_error (const char *reason)
{
  error ("Bad input syntax: %s", reason);
}


static FILE *input;
static Buffer *buffer;
static int ch;			/* the next uneaten charcter */

typedef enum { B_tag, EM, I, PL, S, TT, U, DIGIT, COLOR } Tag_type;

enum { TAG, WHITESPACE, CHAR, END } token; /* the next uneaten token */
int      token_value;
Tag_type token_tag_type;
Flag     token_opening;


static void
nibble (void)
{
  ch = getc (input);
}


static void
gobble1 (int type)
{
  token_tag_type = type;
  token_value = 0;
  nibble ();
}


static void
gobble2 (int type, char expected)
{
  gobble1 (type);
  if (ch != expected)
    syntax_error ("Unexpected character");
  nibble ();
}


static void
eat_tag_name (void)
{
  switch (ch)
    {
    case 'B':      gobble1 (B_tag); break;
    case 'E':      gobble2 (EM, 'M'); break;
    case 'I':      gobble1 (I); break;
    case 'P':      gobble2 (PL, 'L'); break;
    case 'S':      gobble1 (S); break;
    case 'T':      gobble2 (TT, 'T'); break;
    case 'U':      gobble1 (U); break;

    case '0':
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9':
      token_tag_type = DIGIT;
      token_value = ch - '0';
      nibble ();
      break;

    case 'r':
    case 'g':
    case 'b':
    case 'c':
    case 'm':
    case 'y':
    case 'k':
    case 'w':
      token_tag_type = COLOR; 
      token_value = ch; 
      nibble (); 
      break;

    default:
      syntax_error ("Tag name expected");
      break;
    }
}


static void
eat (void)
{
  token = token_tag_type = token_value = token_opening = -1;

  switch (ch)
    {
    case '<': 
      token = TAG;
      nibble ();
      if (ch == '/')
	{
	  token_opening = false;
	  nibble ();
	}
      else
	token_opening = true;
      eat_tag_name ();
      if (ch != '>')
	syntax_error ("'>' expected");
      nibble ();
      break;

    case '\r':
    case '\n':
    case '\t':
    case ' ':
      token = WHITESPACE;
      token_value = ch;
      nibble ();
      break;

    case '>':
      syntax_error ("Unexpected '>'");
      break;

    case EOF:
      token = END;
      break;

    default:
      /* TODO: check here if it's printable? */
      token = CHAR;
      token_value = ch;
      nibble ();
      break;
    }
}


static Context
new_context (Context context)
{
  Context result = context;

  switch (token_tag_type)
    {
    case B_tag:
      result.b = 1;
      break;

    case EM:
      if (!result.s)
	result.em = !result.em;
      break;

    case I:
      result.i = 1;
      break;

    case PL:
      result.s = result.em = result.i = result.b = result.tt = 0;
      result.u = 0;
      break;

    case S:
      result.s = 1;
      result.em = 0;
      break;

    case TT:
      result.tt = 1;
      break;

    case U:
      if (result.u < 3)
	result.u++;
      break;

    case DIGIT:
      result.size = token_value;
      break;

    case COLOR:
      switch (token_value)
	{
	case 'r': result.color = R; break;
	case 'g': result.color = G; break;
	case 'b': result.color = B_color; break;
	case 'c': result.color = C; break;
	case 'm': result.color = M; break;
	case 'y': result.color = Y; break;
	case 'k': result.color = K; break;
	case 'w': result.color = W; break;
	default: assert (false);
	}
      break;
	  
    default:
      printf ("%d\n", token_tag_type);
      assert (false);
    }

  check_context (result);
  return result;
}


static void
parse (Context context)
{
  for (;;)
    switch (token)
      {
      case TAG:
	if (!token_opening)
	  return;
	{
	  Context subcontext = new_context (context);
	  Tag_type tt = token_tag_type;
	  int tv = token_value;
	  eat ();
	  parse (subcontext);
	  if (token != TAG || 
	      token_tag_type != tt || token_value != tv ||
	      token_opening)
	    syntax_error ("Mismatched closing tag");
	  eat ();
	  break;
	}

      case WHITESPACE:
	append_whitespace_char (buffer, context, token_value);
	eat ();
	break;

      case CHAR: 
	append (buffer, context, token_value);
	eat ();
	break;

      case END:
	return;
      
      default:
	assert (false);
	break;
      }
}


static void
parse_document (FILE *doc_file, Buffer *the_buffer)
{
  input = doc_file;
  buffer = the_buffer;
  nibble ();
  eat ();
  parse (make_root_context ());
  assert (token == END);
}


/* Main program */

static void
load (Buffer *buffer, const char *filename)
{
  FILE *fp = open_file (filename, "r");
  parse_document (fp, buffer);
  fclose (fp);
}


int 
main (int argc, char **argv)
{
  static Buffer b1, b2;
  load (&b1, argv[1]);
  load (&b2, argv[2]);
  return equal (&b1, &b2) ? EXIT_SUCCESS : EXIT_FAILURE;
}
